@@ -1,14 +1,14 @@
 {-# LANGUAGE OverloadedStrings #-}
+
 -- | This module contain SQLite-specific functions.
 --
--- /Since: 2.2.8/
+-- @since 2.2.8
 module Database.Esqueleto.SQLite
   ( random_
   ) where
 
-import Database.Esqueleto.Internal.Language hiding (random_)
+import Database.Esqueleto.Internal.Internal hiding (random_)
 import Database.Esqueleto.Internal.PersistentImport
-import Database.Esqueleto.Internal.Sql
 
 -- | (@random()@) Split out into database specific modules
 -- because MySQL uses `rand()`.
